"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var _ = require("lodash");
var Type_1 = require("../Type");
var Strings_1 = require("../Strings");
var Support_1 = require("../Support");
var Source_1 = require("../Source");
var Naming_1 = require("../Naming");
var ConvenienceRenderer_1 = require("../ConvenienceRenderer");
var TargetLanguage_1 = require("../TargetLanguage");
var RendererOptions_1 = require("../RendererOptions");
var unicode = require("unicode-properties");
var L = /** @class */ (function (_super) {
    __extends(L, _super);
    function L() {
        var _this = _super.call(this, "TypeScript", ["typescript", "ts", "tsx"], "ts") || this;
        _this._justTypes = new RendererOptions_1.BooleanOption("just-types", "Interfaces only", false);
        _this._declareUnions = new RendererOptions_1.BooleanOption("explicit-unions", "Explicitly name unions", false);
        _this._omitRuntimeTypecheck = new RendererOptions_1.BooleanOption("no-runtime-typecheck", "Don't verify JSON.parse results at runtime", false);
        _this.setOptions([_this._justTypes, _this._declareUnions, _this._omitRuntimeTypecheck]);
        return _this;
    }
    Object.defineProperty(L.prototype, "supportsOptionalClassProperties", {
        get: function () {
            return true;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(L.prototype, "rendererClass", {
        get: function () {
            return TypeScriptRenderer;
        },
        enumerable: true,
        configurable: true
    });
    return L;
}(TargetLanguage_1.TargetLanguage));
exports.default = L;
function isStartCharacter(utf16Unit) {
    return unicode.isAlphabetic(utf16Unit) || utf16Unit === 0x5f; // underscore
}
function isPartCharacter(utf16Unit) {
    var category = unicode.getCategory(utf16Unit);
    return _.includes(["Nd", "Pc", "Mn", "Mc"], category) || isStartCharacter(utf16Unit);
}
var legalizeName = Strings_1.utf16LegalizeCharacters(isPartCharacter);
function typeNameStyle(original) {
    var words = Strings_1.splitIntoWords(original);
    return Strings_1.combineWords(words, legalizeName, Strings_1.firstUpperWordStyle, Strings_1.firstUpperWordStyle, Strings_1.allUpperWordStyle, Strings_1.allUpperWordStyle, "", isStartCharacter);
}
function propertyNameStyle(original) {
    var escaped = Strings_1.utf16StringEscape(original);
    var quoted = "\"" + escaped + "\"";
    if (original.length === 0) {
        return quoted;
    }
    else if (!isStartCharacter(original.codePointAt(0))) {
        return quoted;
    }
    else if (escaped !== original) {
        return quoted;
    }
    else if (legalizeName(original) !== original) {
        return quoted;
    }
    else {
        return original;
    }
}
var TypeScriptRenderer = /** @class */ (function (_super) {
    __extends(TypeScriptRenderer, _super);
    function TypeScriptRenderer(graph, leadingComments, _justTypes, declareUnions, _omitRuntimeTypecheck) {
        var _this = _super.call(this, graph, leadingComments) || this;
        _this._justTypes = _justTypes;
        _this._omitRuntimeTypecheck = _omitRuntimeTypecheck;
        _this.emitEnum = function (e, enumName) {
            _this.emitBlock(["export enum ", enumName], "", function () {
                _this.forEachEnumCase(e, "none", function (name, jsonName) {
                    _this.emitLine(name, " = \"" + Strings_1.utf16StringEscape(jsonName) + "\",");
                });
            });
        };
        _this.typeMapTypeFor = function (t) {
            return Type_1.matchType(t, function (_anyType) { return "undefined"; }, function (_nullType) { return "null"; }, function (_boolType) { return "false"; }, function (_integerType) { return "0"; }, function (_doubleType) { return "3.14"; }, function (_stringType) { return "\"\""; }, function (arrayType) { return ["A(", _this.typeMapTypeFor(arrayType.items), ")"]; }, function (classType) { return ['O("', _this.nameForNamedType(classType), '")']; }, function (mapType) { return ["M(", _this.typeMapTypeFor(mapType.values), ")"]; }, function (enumType) { return ['E("', _this.nameForNamedType(enumType), '")']; }, function (unionType) {
                var children = unionType.children.map(_this.typeMapTypeFor);
                return ["U("].concat(Support_1.intercalate(", ", children).toArray(), [")"]);
            });
        };
        _this.emitBlock = function (source, end, emit) {
            _this.emitLine(source, " {");
            _this.indent(emit);
            _this.emitLine("}", end);
        };
        _this.emitTypeMap = function () {
            _this.emitBlock("const typeMap: any =", ";", function () {
                _this.forEachClass("none", function (t, name) {
                    _this.emitBlock(['"', name, '":'], ",", function () {
                        _this.forEachClassProperty(t, "none", function (propName, _propJsonName, property) {
                            _this.emitLine(propName, ": ", _this.typeMapTypeForProperty(property), ",");
                        });
                    });
                });
                _this.forEachEnum("none", function (e, name) {
                    _this.emitLine('"', name, '": [');
                    _this.indent(function () {
                        _this.forEachEnumCase(e, "none", function (caseName) {
                            _this.emitLine(name, ".", caseName, ",");
                        });
                    });
                    _this.emitLine("],");
                });
            });
        };
        _this.emitClass = function (c, className) {
            _this.emitBlock(["export interface ", className], "", function () {
                var table = [];
                _this.forEachClassProperty(c, "none", function (name, _jsonName, p) {
                    var t = p.type;
                    var nullable = t instanceof Type_1.UnionType ? Type_1.nullableFromUnion(t) : null;
                    if (p.isOptional && nullable === null) {
                        nullable = t;
                    }
                    table.push([
                        [name, nullable !== null ? "?" : "", ": "],
                        [_this.sourceFor(nullable !== null ? nullable : t).source, ";"]
                    ]);
                });
                _this.emitTable(table);
            });
        };
        _this.emitConvertModule = function () {
            _this.emitMultiline("// Converts JSON strings to/from your types");
            if (!_this._omitRuntimeTypecheck) {
                _this.emitMultiline("// and asserts the results of JSON.parse at runtime");
            }
            _this.emitBlock("export module Convert", "", function () {
                _this.forEachTopLevel("interposing", function (t, name) {
                    _this.emitBlock(["export function to", name, "(json: string): ", _this.sourceFor(t).source], "", function () {
                        if (_this._omitRuntimeTypecheck) {
                            _this.emitLine("return JSON.parse(json);");
                        }
                        else {
                            _this.emitLine("return cast(JSON.parse(json), ", _this.typeMapTypeFor(t), ");");
                        }
                    });
                    _this.ensureBlankLine();
                    var camelCaseName = Source_1.modifySource(Strings_1.camelCase, name);
                    _this.emitBlock(["export function ", camelCaseName, "ToJson(value: ", _this.sourceFor(t).source, "): string"], "", function () {
                        _this.emitLine("return JSON.stringify(value, null, 2);");
                    });
                });
                if (!_this._omitRuntimeTypecheck) {
                    _this.emitMultiline("\nfunction cast<T>(obj: any, typ: any): T {\n    if (!isValid(typ, obj)) {\n        throw `Invalid value`;\n    }\n    return obj;\n}\n\nfunction isValid(typ: any, val: any): boolean {\n    if (typ === undefined) return true;\n    if (typ === null) return val === null || val === undefined;\n    return typ.isUnion  ? isValidUnion(typ.typs, val)\n            : typ.isArray  ? isValidArray(typ.typ, val)\n            : typ.isMap    ? isValidMap(typ.typ, val)\n            : typ.isEnum   ? isValidEnum(typ.name, val)\n            : typ.isObject ? isValidObject(typ.cls, val)\n            :                isValidPrimitive(typ, val);\n}\n\nfunction isValidPrimitive(typ: string, val: any) {\n    return typeof typ === typeof val;\n}\n\nfunction isValidUnion(typs: any[], val: any): boolean {\n    // val must validate against one typ in typs\n    return typs.find(typ => isValid(typ, val)) !== undefined;\n}\n\nfunction isValidEnum(enumName: string, val: any): boolean {\n    const cases = typeMap[enumName];\n    return cases.indexOf(val) !== -1;\n}\n\nfunction isValidArray(typ: any, val: any): boolean {\n    // val must be an array with no invalid elements\n    return Array.isArray(val) && val.every((element, i) => {\n        return isValid(typ, element);\n    });\n}\n\nfunction isValidMap(typ: any, val: any): boolean {\n    if (val === null || typeof val !== \"object\" || Array.isArray(val)) return false;\n    // all values in the map must be typ\n    return Object.keys(val).every(prop => {\n        if (!Object.prototype.hasOwnProperty.call(val, prop)) return true;\n        return isValid(typ, val[prop]);\n    });\n}\n\nfunction isValidObject(className: string, val: any): boolean {\n    if (val === null || typeof val !== \"object\" || Array.isArray(val)) return false;\n    let typeRep = typeMap[className];\n    return Object.keys(typeRep).every(prop => {\n        if (!Object.prototype.hasOwnProperty.call(typeRep, prop)) return true;\n        return isValid(typeRep[prop], val[prop]);\n    });\n}\n\nfunction A(typ: any) {\n    return { typ, isArray: true };\n}\n\nfunction E(name: string) {\n    return { name, isEnum: true };\n}\n\nfunction U(...typs: any[]) {\n    return { typs, isUnion: true };\n}\n\nfunction M(typ: any) {\n    return { typ, isMap: true };\n}\n\nfunction O(className: string) {\n    return { cls: className, isObject: true };\n}\n");
                    _this.emitTypeMap();
                }
            });
        };
        _this.emitUnion = function (u, unionName) {
            if (_this._inlineUnions) {
                return;
            }
            var children = Source_1.multiWord.apply(void 0, [" | "].concat(u.children.map(function (c) { return Source_1.parenIfNeeded(_this.sourceFor(c)); }).toArray()));
            _this.emitLine("export type ", unionName, " = ", children.source, ";");
        };
        _this._inlineUnions = !declareUnions;
        return _this;
    }
    TypeScriptRenderer.prototype.topLevelNameStyle = function (rawName) {
        return typeNameStyle(rawName);
    };
    TypeScriptRenderer.prototype.makeNamedTypeNamer = function () {
        return new Naming_1.Namer("types", typeNameStyle, []);
    };
    TypeScriptRenderer.prototype.namerForClassProperty = function () {
        return new Naming_1.Namer("properties", propertyNameStyle, []);
    };
    TypeScriptRenderer.prototype.makeUnionMemberNamer = function () {
        return null;
    };
    TypeScriptRenderer.prototype.makeEnumCaseNamer = function () {
        return new Naming_1.Namer("enum-cases", typeNameStyle, []);
    };
    TypeScriptRenderer.prototype.namedTypeToNameForTopLevel = function (type) {
        return Type_1.directlyReachableSingleNamedType(type);
    };
    TypeScriptRenderer.prototype.sourceFor = function (t) {
        var _this = this;
        return Type_1.matchType(t, function (_anyType) { return Source_1.singleWord("any"); }, function (_nullType) { return Source_1.singleWord("null"); }, function (_boolType) { return Source_1.singleWord("boolean"); }, function (_integerType) { return Source_1.singleWord("number"); }, function (_doubleType) { return Source_1.singleWord("number"); }, function (_stringType) { return Source_1.singleWord("string"); }, function (arrayType) {
            var itemType = _this.sourceFor(arrayType.items);
            if ((arrayType.items instanceof Type_1.UnionType && _this._inlineUnions) ||
                arrayType.items instanceof Type_1.ArrayType) {
                return Source_1.singleWord(["Array<", itemType.source, ">"]);
            }
            else {
                return Source_1.singleWord([Source_1.parenIfNeeded(itemType), "[]"]);
            }
        }, function (classType) { return Source_1.singleWord(_this.nameForNamedType(classType)); }, function (mapType) { return Source_1.singleWord(["{ [key: string]: ", _this.sourceFor(mapType.values).source, " }"]); }, function (enumType) { return Source_1.singleWord(_this.nameForNamedType(enumType)); }, function (unionType) {
            if (_this._inlineUnions || Type_1.nullableFromUnion(unionType) !== null) {
                var children = unionType.children.map(function (c) { return Source_1.parenIfNeeded(_this.sourceFor(c)); });
                return Source_1.multiWord.apply(void 0, [" | "].concat(children.toArray()));
            }
            else {
                return Source_1.singleWord(_this.nameForNamedType(unionType));
            }
        });
    };
    TypeScriptRenderer.prototype.typeMapTypeForProperty = function (p) {
        if (!p.isOptional || p.type.isNullable) {
            return this.typeMapTypeFor(p.type);
        }
        return ["U(null, ", this.typeMapTypeFor(p.type), ")"];
    };
    TypeScriptRenderer.prototype.emitSourceStructure = function () {
        var _this = this;
        if (this.leadingComments !== undefined) {
            this.emitCommentLines("// ", this.leadingComments);
        }
        else if (!this._justTypes) {
            this.emitMultiline("// To parse this data:\n//");
            var topLevelNames_1 = [];
            this.forEachTopLevel("none", function (_t, name) {
                topLevelNames_1.push(", ", name);
            }, Type_1.isNamedType);
            this.emitLine("//   import { Convert", topLevelNames_1, ' } from "./file";');
            this.emitLine("//");
            this.forEachTopLevel("none", function (_t, name) {
                var camelCaseName = Source_1.modifySource(Strings_1.camelCase, name);
                _this.emitLine("//   const ", camelCaseName, " = Convert.to", name, "(json);");
            });
            if (!this._omitRuntimeTypecheck) {
                this.emitLine("//");
                this.emitLine("// These functions will throw an error if the JSON doesn't");
                this.emitLine("// match the expected interface, even if the JSON is valid.");
            }
        }
        this.forEachNamedType("leading-and-interposing", this.emitClass, this.emitEnum, this.emitUnion);
        if (!this._justTypes) {
            this.ensureBlankLine();
            this.emitConvertModule();
        }
    };
    return TypeScriptRenderer;
}(ConvenienceRenderer_1.ConvenienceRenderer));
