import { Map, Set, OrderedSet, Collection } from "immutable";
import { Type, SeparatedNamedTypes } from "./Type";
import { GraphRewriteBuilder, TypeRef, TypeBuilder, StringTypeMapping } from "./TypeBuilder";
import { Graph } from "./Graph";
import { TypeAttributeKind, TypeAttributes } from "./TypeAttributes";
export declare class TypeAttributeStore {
    private readonly _typeGraph;
    private _values;
    private _topLevelValues;
    constructor(_typeGraph: TypeGraph, _values: (TypeAttributes | undefined)[]);
    private getTypeIndex(t);
    attributesForType(t: Type): TypeAttributes;
    attributesForTopLevel(name: string): TypeAttributes;
    private setInMap<T>(attributes, kind, value);
    set<T>(kind: TypeAttributeKind<T>, t: Type, value: T): void;
    setForTopLevel<T>(kind: TypeAttributeKind<T>, topLevelName: string, value: T): void;
    private tryGetInMap<T>(attributes, kind);
    tryGet<T>(kind: TypeAttributeKind<T>, t: Type): T | undefined;
    tryGetForTopLevel<T>(kind: TypeAttributeKind<T>, topLevelName: string): T | undefined;
}
export declare class TypeAttributeStoreView<T> {
    private readonly _attributeStore;
    private readonly _definition;
    constructor(_attributeStore: TypeAttributeStore, _definition: TypeAttributeKind<T>);
    set(t: Type, value: T): void;
    setForTopLevel(name: string, value: T): void;
    tryGet(t: Type): T | undefined;
    get(t: Type): T;
    tryGetForTopLevel(name: string): T | undefined;
    getForTopLevel(name: string): T;
}
export declare class TypeGraph {
    private _typeBuilder?;
    private _attributeStore;
    private _topLevels?;
    private _types?;
    constructor(typeBuilder: TypeBuilder);
    private readonly isFrozen;
    readonly attributeStore: TypeAttributeStore;
    freeze(topLevels: Map<string, TypeRef>, types: Type[], typeAttributes: (TypeAttributes | undefined)[]): void;
    readonly topLevels: Map<string, Type>;
    atIndex(index: number): [Type, TypeAttributes];
    filterTypes(predicate: ((t: Type) => boolean) | undefined, childrenOfType: ((t: Type) => Collection<any, Type>) | undefined, topDown: boolean): OrderedSet<Type>;
    allNamedTypes: (childrenOfType?: ((t: Type) => Collection<any, Type>) | undefined) => OrderedSet<Type>;
    allNamedTypesSeparated: (childrenOfType?: ((t: Type) => Collection<any, Type>) | undefined) => SeparatedNamedTypes;
    rewrite<T extends Type>(stringTypeMapping: StringTypeMapping, alphabetizeProperties: boolean, replacementGroups: T[][], replacer: (typesToReplace: Set<T>, builder: GraphRewriteBuilder<T>, forwardingRef: TypeRef) => TypeRef): TypeGraph;
    garbageCollect(alphabetizeProperties: boolean): TypeGraph;
    allTypesUnordered: () => Set<Type>;
    makeGraph(invertDirection: boolean, childrenOfType: (t: Type) => OrderedSet<Type>): Graph<Type>;
}
export declare function noneToAny(graph: TypeGraph, stringTypeMapping: StringTypeMapping): TypeGraph;
export declare function optionalToNullable(graph: TypeGraph, stringTypeMapping: StringTypeMapping): TypeGraph;
